import os
import cv2
import re
import math
import shutil
import random
import numpy as np
import matplotlib.pyplot as plt
from naeural_core import Logger



if __name__ == '__main__':
  path = '/home/work/Pictures'
  l = Logger('YOLOBM', base_folder='.', app_folder='_local_cache')
  path_list = [
               # os.path.join(l.get_dropbox_drive(), '_vapor_data/_lpr_v2/_data/Rares_Efect/Good MULTI 10/Train'),
               # os.path.join(l.get_dropbox_drive(), '_vapor_data/_lpr_v2/_data/Rares_Efect/Good MULTI 10/Dev'),
               os.path.join(l.get_dropbox_drive(), '_vapor_data/_lpr_v2/_data/Rares_Efect/Good MULTI 10/FIX'),
               # os.path.join(l.get_dropbox_drive(), '_vapor_data/_lpr_v2/_data/Rares_Efect/Good MULTI 10/Test')
               ]
  for path in path_list:
    path_good = path + '/GOOD'
    path_edited = path + '/EDITED'
    shutil.rmtree(path_edited, ignore_errors=True)
    if not os.path.exists(path_edited):
      os.makedirs(path_edited)
    for img_name in os.listdir(path_good):
      img_path = path_good + '/' + img_name
      img_edited_path = path_edited + '/' + img_name
      img = cv2.imread(img_path)
      height, width = img.shape[:2]
      X_TRANSLATE = width / 100
      Y_TRANSLATE = height / 100
      # radius = random.uniform(width /100, width / 20)
      radius = width / 90
      sign = random.uniform(-1, 1)
      X_TRANSLATE = int(random.uniform(0, radius))
      if sign < 0:
        X_TRANSLATE = -X_TRANSLATE
        
      
      sign = random.uniform(-1, 1)
      Y_TRANSLATE = int(math.sqrt(math.pow(radius, 2) - math.pow(X_TRANSLATE, 2)))
      if sign < 0:
        Y_TRANSLATE = -Y_TRANSLATE
      img_R = img[:,:,2]
      img_G = img[:,:,1]
      img_B = img[:,:,0]
      Translate_G = np.float32([[1, 0, X_TRANSLATE], [0, 1, Y_TRANSLATE]])
      Translate_B = np.float32([[1, 0, 2*X_TRANSLATE], [0, 1, 2*Y_TRANSLATE]])
      img_translation_G = cv2.warpAffine(img_G, Translate_G, (width, height))
      img_translation_B = cv2.warpAffine(img_B, Translate_B, (width, height))
      final_img = cv2.merge((img_R, img_translation_G, img_translation_B))
    
      if X_TRANSLATE > 0:
        final_img = final_img[:,2 * X_TRANSLATE:,:]
      elif X_TRANSLATE < 0:
        final_img = final_img[:,:2 * X_TRANSLATE,:]
        
      if Y_TRANSLATE > 0:
        final_img = final_img[2 * Y_TRANSLATE:,:,:]
      elif Y_TRANSLATE < 0:
        final_img = final_img[:2 * Y_TRANSLATE,:,:]
        
      final_img = cv2.resize(final_img, (width, height))
        
      
      cv2.imwrite(img_edited_path, final_img)




import os


def is_image(file_path):
  return file_path.lower().endswith(('.jpg', '.jpeg', 'png', 'bmp'))


def dir_path_to_dict(path, exclude_dirs=None):
  exclude_dirs = exclude_dirs or []
  files_dict = {}
  for subdir, dirs, files in os.walk(path):
    for file in files:
      curr_path = os.path.join(subdir, file)
      sub_ds_name = os.path.split(subdir)[-1]
      if sub_ds_name in exclude_dirs:
        continue
      if sub_ds_name not in files_dict.keys():
        files_dict[sub_ds_name] = []
      if 'pickle' not in curr_path and is_image(curr_path):
        files_dict[sub_ds_name].append(curr_path)
    # endfor file
  # endfor subdir
  return files_dict

keep_dirs = None
test_ds_path = 'C:\\resources\\all_servings_tests'
files_dict = dir_path_to_dict(test_ds_path, exclude_dirs=keep_dirs)


if __name__ == '__main__':
  print(files_dict)



import gensim
import gensim.downloader as api

from naeural_core import Logger

if __name__ == '__main__':
  l = Logger('WORDS', base_folder='.', app_folder='_local_cache')
  l.P("loading dict...")
  fn = 'test.pkl'
  if l.get_data_file('test.pkl') is not None:
    words = l.load_pickle_from_data(fn)
  else:
    wv = api.load('word2vec-google-news-300')
    words = wv.index_to_key
    l.save_pickle_to_data(data=words, fn=fn)
  
  WORD_SIZE = 5
  excluded = "eropadcnm"
  included = {
    'i' : 1,
    's' : 4,
    't' : {
      'not' : [0,2],
      },
    }

  checked = []
  l.p("processing...")
  for i, word in enumerate(words):
    word = word.lower()
    if len(word) != WORD_SIZE:
      continue
    if word in checked:
      continue
    checked.append(word)
    found = True
    for excl in excluded:
      if excl in word:
        found = False
        break
    if not found:
      continue
    for incl in included:
      if isinstance(included[incl], int):
        if word[included[incl]] != incl:
          found = False
          break
      else:
        for pos in included[incl]['not']:
          if word[pos] == incl:
            found = False
            break
      if not found:
        break
    if not found:
      continue
    print(word, flush=True, end=' ')



from naeural_core import constants as ct

from naeural_core.data.mixins_libs import _SqlQueryAcquisitionMixin
from naeural_core.data.base import DataCaptureThread

from dateutil.relativedelta import relativedelta
from datetime import datetime

_CONFIG = {
  **DataCaptureThread.CONFIG,
  
  'SAVE_LAST_STATE' : True,

  'VALIDATION_RULES' : {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
    'SAVE_LAST_STATE': {
      'TYPE': 'bool'
    }
  },


}


class SqlQueryDataCapture(DataCaptureThread, _SqlQueryAcquisitionMixin):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    self.odbc_connector = None
    super(SqlQueryDataCapture, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    if not self.cfg_save_last_state:
      self.P("WARNING!!! Sql query DCT is running with `save_last_state` False")

  @property
  def cfg_nr_retries(self):
    return self.cfg_stream_config_metadata.get('NR_RETRIES', 3)

  @property
  def cfg_connector_type(self):
    return self.cfg_stream_config_metadata.get('CONNECTOR_TYPE', 'PYMSSQL')

  def _init(self):
    self._metadata.update(
      # TODO
    )

    self.connection_info = self.cfg_stream_config_metadata['CONNECTION_INFO']
    self.queries = self.cfg_stream_config_metadata.get('QUERY', None)
    self.stored_procedures =  self.cfg_stream_config_metadata.get('STORED_PROCEDURE', None)
    assert self.queries is not None or self.stored_procedures is not None, "A stored procedure or query must be configured"

    if self.stored_procedures is not None and type(self.stored_procedures) is dict:
      self.stored_procedures = [self.stored_procedures]

    if self.queries is not None and type(self.queries) is dict:
      self.queries = [self.queries]

    self._load_state()

    if self.cfg_connector_type == 'PYMSSQL':
      from naeural_core.local_libraries.db_conn.pymssql_conn import PyMSSQLConnector
      config = {
        'CONNECT_PARAMS': {
          **{k.lower():v for k,v in self.connection_info.items()}
        },
      }
      self.conn = PyMSSQLConnector(log=self.log, config=config)
    else:
      from naeural_core.local_libraries.db_conn.odbc_conn import ODBCConnector
      config = {
        'CONNECT_PARAMS': {
          'Encrypt': 'yes',
          'TrustServerCertificate': 'yes',
          'Connection Timeout': 30,
          **self.connection_info
        },
      }
      self.conn = ODBCConnector(log=self.log, config=config)
    #endif

    self.conn.connect(nr_retries=self.cfg_nr_retries)


    # TODO: Check with andrei, to add validation rules / self.cfg_backlog
    if self.cfg_stream_config_metadata.get("BACKLOG") == "FULL":
      self._backlog_start_time = datetime.strftime(datetime(1900, 1, 1, 0, 0), "%Y-%m-%d %H:%M:%S")
    elif self.cfg_stream_config_metadata.get("BACKLOG") == 'YEAR':
      self._backlog_start_time = datetime.strftime(datetime.now() - relativedelta(years=1), "%Y-%m-%d %H:%M:%S")
    elif self.cfg_stream_config_metadata.get("BACKLOG") == 'MONTH':
      self._backlog_start_time = datetime.strftime(datetime.now() - relativedelta(months=1), "%Y-%m-%d %H:%M:%S")
    elif self.cfg_stream_config_metadata.get("BACKLOG") == 'WEEK':
      self._backlog_start_time = datetime.strftime(datetime.now() - relativedelta(weeks=1), "%Y-%m-%d %H:%M:%S")
    elif self.cfg_stream_config_metadata.get("BACKLOG") == 'DAY':
      self._backlog_start_time = datetime.strftime(datetime.now() - relativedelta(days=1), "%Y-%m-%d %H:%M:%S")
    else:
      raise ValueError("Invalid configuration value for 'BACKLOG' : '{}'".format(self.cfg_stream_config_metadata.get("BACKLOG")))
    return

  def _maybe_reconnect(self):
    if self.conn.check_connection(): ### TODO: maybe redo implementation
      return

    self.conn.connect(nr_retries=self.cfg_max_retries)

    if self.conn.check_connection():
      self.P('Done init')
    else:
      msg = 'Data capture could not be initialized after {} retries'.format(self.cfg_max_retries)
      self.P(msg)
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
      )
    return

  def _run_data_aquisition_step(self): ### TODO: refactor typo; it's acquisition
    current_datetime = datetime.now().replace(second=0, microsecond=0)

    # self.P('Preparing sql queries ...')
    sql_queries = self._get_sql_queries(current_datetime)
    if len(sql_queries) == 0:
      return
    
    self.P("Data acquisition step started ...")

    inputs = []
    dct_dfs = {}
    for datasource_name, sql_query in sql_queries.items():
      self.P("Running query for datasource '{}' ...".format(datasource_name))
      query_params = {
          'sql_query' : sql_query,
          'chunksize' : 1000000, ### if removed, then the generator conn.data_chunk_generator() will have only one step
        }
      df_data = self.conn.get_data(**query_params)
      self.P("Query ran successfully")
      self.P("Extracted {} rows".format(len(df_data)))
      if df_data is not None and len(df_data) > 0:
        dct_dfs[datasource_name] = df_data
      #endif
    #endfor

    if len(dct_dfs) > 0:
      inputs.append(
        self._new_input(struct_data=dct_dfs, metadata=self._metadata.__dict__.copy())
      )
      self._add_inputs(inputs)
      self._save_state()
    #endif
    self.P("Data acquisition step finished")
    return



import os

class _BetaInferenceMixin(object):
  """
  Mixin for inference functionalities that are attached to `libraries.logger.Logger` (onnx, trt)

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_BasicPyTorchMixin`:
    - self.load_pytorch_model

  * Obs: This mixin uses also attributes/methods of `_JSONSerializationMixin`:
    - self.load_models_json

  """

  def __init__(self):
    super(_BetaInferenceMixin, self).__init__()

    try:
      from .basic_pytorch_mixin import _BasicPyTorchMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _BetaInferenceMixin without having _BasicPyTorchMixin")

    try:
      from ratio1.logging.logger_mixins.json_serialization_mixin import _JSONSerializationMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _BetaInferenceMixin without having _JSONSerializationMixin")

    return

  ###############################
  ###                         ###
  ###  ONNX area              ###
  ###                         ###
  ###############################
  def load_onnx_model(self, model_name, full_path=False):
    import onnx
    import onnxruntime as ort

    if not model_name.endswith('.onnx'):
      model_name += '.onnx'

    if not full_path:
      model_full_path = os.path.join(self.get_models_folder(), model_name)
    else:
      model_full_path = model_name

    if not os.path.isfile(model_full_path):
      self.p('Provided path does not exists {}'.format(model_full_path))
      return

    self.p('Loading onnx model from: {}'.format(model_full_path))
    model = onnx.load(model_full_path)
    onnx.checker.check_model(model)
    onnx.helper.printable_graph(model.graph)

    ort_session = ort.InferenceSession(model_full_path)

    output_names = [node.name for node in model.graph.output]

    # TODO: review. this was needed for pytorch yolov3/yolov5
    l = []
    for x in output_names:
      try:
        int(x)
      except:
        l.append(x)
    output_names = l

    input_all = [node.name for node in model.graph.input]
    input_initializer = [node.name for node in model.graph.initializer]
    input_names = list(set(input_all) - set(input_initializer))

    return model, ort_session, input_names, output_names

  def saved_keras_to_onnx(self, model, onnx_name, folder='models'):
    import keras2onnx as k2o
    assert folder in [None, 'data', 'output', 'models']

    onnx_model = k2o.convert_keras(
      model=model,
      name=onnx_name
    )

    if not onnx_name.endswith('.onnx'):
      onnx_name += '.onnx'

    lfld = self.get_target_folder(target=folder)

    if folder is not None:
      path = os.path.join(lfld, onnx_name)
    else:
      path = onnx_name

    self.p('Saving keras model to onnx model at: {}'.format(path))
    k2o.save_model(onnx_model, path)
    self.p('Done saving onnx model')
    return

  def saved_pytorch_model_to_onnx(self, model_name,
                                  file_name,
                                  input_names=None,
                                  output_names=None,
                                  batch_size=None,
                                  input_height=None,
                                  input_width=None,
                                  input_channels=3,
                                  opset_version=11
                                  ):
    model = self.load_pytorch_model(model_name)
    path = self.pytorch_to_onnx(
      model=model,
      file_name=file_name,
      input_names=input_names,
      output_names=output_names,
      batch_size=batch_size,
      input_height=input_height,
      input_width=input_width,
      input_channels=input_channels,
      opset_version=opset_version
    )
    return path

  def pytorch_to_onnx(self, model,
                      file_name,
                      input_names=None,
                      output_names=None,
                      batch_size=None,
                      input_height=None,
                      input_width=None,
                      input_channels=3,
                      opset_version=11
                      ):
    import torch as th
    DEVICE = th.device('cuda:0' if th.cuda.is_available() else 'cpu')

    self.p('Converting {} to onnx'.format(model.__class__.__name__))
    assert (not input_height and not input_width) or (input_height and input_width)

    model.eval()
    EXT = '.pth.onnx'
    if not file_name.endswith(EXT):
      file_name = file_name + EXT

    input_shape, is_dynamic_shape, is_dynamic_batch = None, False, False
    if not batch_size and not input_height and not input_width:
      is_dynamic_shape = True
      input_shape = (1, input_channels, 1080, 1920)  # random numbera for input shape
    elif not batch_size:
      is_dynamic_batch = True
      input_shape = (
      1, input_channels, input_height, input_width)  # random number for batch_size. the batch_size will be made dynamic

    if not input_names:
      input_names = ['input']
    if not output_names:
      output_names = ['output']

    if is_dynamic_shape:
      dynamic_axes = {
        'input': {0: 'batch_size', 2: 'height', 3: 'width'},
        'output': {0: 'batch_size', 2: 'height', 3: 'width'}
      }  # adding names for better debugging
    elif is_dynamic_batch:
      dynamic_axes = {
        'input': {0: 'batch_size'},
        'output': {0: 'batch_size'}
      }
    else:
      dynamic_axes = None

    save_path = os.path.join(self.get_models_folder(), file_name)
    self.p('Saving model to: {}'.format(save_path[-50:]))

    th_inputs = th.zeros(*input_shape).to(DEVICE)
    th.onnx.export(
      model=model,
      args=th_inputs,
      f=save_path,
      input_names=input_names,
      output_names=output_names,
      dynamic_axes=dynamic_axes,
      opset_version=opset_version,
      verbose=True
    )

    self.p('Done converting pytorch model to onnx', show_time=True)
    return save_path

  ### end ONNX area

  ###############################
  ###                         ###
  ###  TRT  area              ###
  ###                         ###
  ###############################
  def load_tensorflow_trt_graph(self, graph_file, folder='models'):
    from ..nn.tf.utils import load_tf_graph, load_tf_graph_saved_model
    assert folder in [None, 'models']

    if not graph_file.endswith('.trt'):
      graph_file += '.trt'
    if folder is None:
      path = graph_file
    else:
      path = os.path.join(self.get_models_folder(), graph_file)

    if os.path.exists(path):
      cfg = self.load_models_json(graph_file + '.txt')
      lst_inputs = [v for k, v in cfg.items() if k.startswith('INPUT')]
      lst_outputs = [v for k, v in cfg.items() if k.startswith('OUTPUT')]
      try:
        self.p('Try loading graph is standard load graph method')
        trt_graph = load_tf_graph(log=self, pb_file=path, return_elements=lst_outputs)
      except:
        self.p('Could not load graph with standard load method.')
        self.p('Try loading graph from a saved model checkpoint')
        trt_graph = load_tf_graph_saved_model(log=self, pb_file=path)
      return trt_graph, lst_inputs, lst_outputs
    else:
      self.p('TRT graph not found on location: {}'.format(path))
    return

  def load_pytorch_trt_graph(self, graph_file, folder='models'):
    import torch as th
    from torch2trt import TRTModule
    assert folder in [None, 'models']

    if not graph_file.endswith('.trt'):
      graph_file += '.trt'
    if folder is None:
      path = graph_file
    else:
      path = os.path.join(self.get_models_folder(), graph_file)
    if os.path.exists(path):
      trt_graph = TRTModule()
      trt_graph.load_state_dict(th.load(path))
      # TODO: find a better way to obtain pytorch input/output names
      lst_inputs, lst_outputs = ['input'], ['output']
      return trt_graph, lst_inputs, lst_outputs
    else:
      self.p('TRT graph not found on location: {}'.format(path))
    return

  def pytorch_to_trt(self, model,
                     file_name,
                     input_height,
                     input_width,
                     input_channels,
                     max_batch_size=30):
    import torch as th
    from torch2trt import torch2trt

    DEVICE = th.device('cuda:0' if th.cuda.is_available() else 'cpu')

    EXT = '.pth.trt'
    if not file_name.endswith(EXT):
      file_name += EXT

    # create dummy input tensor
    th_x = th.ones((1, input_channels, input_height, input_width)).to(DEVICE)

    # convert to TensorRT feeding sample data as input
    # TODO: extend options for model converter
    model_trt = torch2trt(
      module=model,
      inputs=[th_x],
      max_batch_size=max_batch_size
    )

    path_model = os.path.join(self.get_models_folder(), file_name)
    th.save(model_trt.state_dict(), path_model)

    # load trt model for checking
    self.load_pytorch_trt_graph(graph_file=file_name)
    return file_name

  def tensorflow_graph_to_trt(self, graph_name,
                              input_names=None,
                              output_names=None,
                              folder='models',
                              max_batch_size=30
                              ):
    import tensorflow.compat.v1 as tf
    from ..nn.tf.utils import save_graph_to_file
    tf.disable_eager_execution()
    from tensorflow.python.compiler.tensorrt import trt_convert as trt
    EXT = '.trt'

    assert folder in [None, 'models']
    if folder:
      path_graph = self.get_models_file(graph_name)
      cfg = self.load_models_json(graph_name + '.txt')
      input_names = [v for k, v in cfg.items() if k.startswith('INPUT_')]
      output_names = [v for k, v in cfg.items() if k.startswith('OUTPUT_')]
      input_names = [x.split(':')[0] for x in input_names]
      output_names = [x.split(':')[0] for x in output_names]
    else:
      path_graph = graph_name
      assert input_names and output_names, 'Please provide input and output graph names'

    assert os.path.isfile(path_graph), 'Graph not found in {}'.format(path_graph)

    with tf.Session() as sess:
      # First deserialize the frozen graph
      with tf.gfile.GFile(path_graph, 'rb') as f:
        frozen_graph = tf.GraphDef()
        frozen_graph.ParseFromString(f.read())
      # endwith

      # Now you can create a TensorRT inference graph from frozen graph
      # TODO: extend options for model converter
      converter = trt.TrtGraphConverter(
        input_graph_def=frozen_graph,
        nodes_blacklist=output_names,
        max_batch_size=max_batch_size
      )  # output nodes
      trt_graph = converter.convert()

      trt_graph_name = graph_name + EXT

      save_graph_to_file(
        log=self,
        sess=sess,
        graph=trt_graph,
        pb_file=trt_graph_name,
        input_tensor_names=input_names,
        output_tensor_names=output_names
      )
    return trt_graph
  ### end TRT area


"""
  ⚠️ do not modify this ⚠️
"""

import multiprocessing as mp
import os
import sys
import warnings

warnings.filterwarnings("ignore")

from naeural_core.main.entrypoint import main
  
  
if __name__ == '__main__':  
  mp.set_start_method('spawn') # if moved at import will generate errors in subprocs
  exit_code, eng = main()
  
  # TODO: configured with flag in startup
  SYS_EXIT = False 
  
  if SYS_EXIT:
    print("Executing sys.exit({})...".format(exit_code))
    sys.exit(exit_code)
  else:
    print("Executing os._exit({})...".format(exit_code))
    os._exit(exit_code)
  #endif sys exit
  



from naeural_core import Logger
from threading import Thread
from time import time
import os

if __name__ == '__main__':

  log = Logger(lib_name='tstp', base_folder='.', app_folder='_local_cache')

  # tested on local files (summing 500mb), use your own file if needed
  fld = '/Users/laurentiupiciu/Downloads/'
  files = ['output_part_000.mkv', 'output_part_001.mkv', 'output_part_002.mkv']

  threads = []

  endpoint = 'MINIO_ENDPOINT'
  access_key = 'MINIO_ACCESS_KEY'
  secret_key = 'MINIO_SECRET_KEY'
  bucket_name ="MINIO_BUCKET_NAME"

  if True:
    # UPLOAD TEST
    start_parallel_up = time()
    for fn in files:
      path = os.path.join(fld, fn)
      t = Thread(target=log.minio_upload, args=(path, endpoint, access_key, secret_key, bucket_name), daemon=True)
      threads.append(t)
      t.start()

    for t in threads:
      t.join()

    """
    # Logs from `log.minio_upload` for each thread
    #  * Uploaded '/Users/laurentiupiciu/Downloads/output_part_001.mkv' as '...' in 13.17s
    #  * Uploaded '/Users/laurentiupiciu/Downloads/output_part_000.mkv' as '...' in 13.18s
    #  * Uploaded '/Users/laurentiupiciu/Downloads/output_part_002.mkv' as '...' in 13.22s
    """
    end_parallel_up = time()

    urls = []
    start_seq_up = time()
    for fn in files:
      path = os.path.join(fld, fn)
      res = log.minio_upload(
        file_path=path, endpoint=endpoint, access_key=access_key,
        secret_key=secret_key, bucket_name=bucket_name
      )
      urls.append(res)
    #endfor

    """
    # Logs for `log.minio_upload` for each call:
    #  * Uploaded '/Users/laurentiupiciu/Downloads/output_part_000.mkv' as '...' in 3.99s
    #  * Uploaded '/Users/laurentiupiciu/Downloads/output_part_001.mkv' as '...' in 5.85s
    #  * Uploaded '/Users/laurentiupiciu/Downloads/output_part_002.mkv' as '...' in 4.61s
    """
    end_seq_up = time()

    log.P("Total time parallel   upload: {:.2f}s".format(end_parallel_up-start_parallel_up))
    log.P("Total time sequential upload: {:.2f}s".format(end_seq_up-start_seq_up))

    """
    [tstp][2022-05-31 18:21:47] Total time parallel   upload: 13.33s
    [tstp][2022-05-31 18:21:47] Total time sequential upload: 14.45s
    """
    print(urls)
  #endif




#global dependencies
import os

#local dependencies
from naeural_core import constants as ct

from naeural_core import Logger
from naeural_core.utils.config_utils import get_config_from_code

if __name__ == '__main__':
  log = Logger(
    lib_name='GPC', 
    base_folder='.',
    app_folder='_local_cache',
    config_file='config_startup.txt', 
    TF_KERAS=False
    )  
  
  path = ct.PLUGIN_SEARCH.LOC_BIZ_PLUGINS.replace('.', '/')
  files = [x for x in os.listdir(path) if x.endswith('.py')]
  files = [os.path.join(path, x) for x in files]
  dct = {}
  for fn in files:
    cfg = get_config_from_code(fn, log)
    if cfg is not None:
      signature = os.path.basename(fn).split('.')[0].upper()
      cfg[ct.SIGNATURE] = signature
      dct[signature] = cfg

  fn_out = '{}_all_plugins_config.txt'.format(log.now_str())
  log.save_output_json(
    data_json=dct, 
    fname=fn_out
    )



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Common modules
"""

import logging
import math
import warnings
from copy import copy
from pathlib import Path

import numpy as np
import pandas as pd
import requests
import torch
import torch.nn as nn
from PIL import Image
from torch.cuda import amp

from y5utils.datasets import exif_transpose, letterbox
from y5utils.general import colorstr, increment_path, make_divisible, non_max_suppression, save_one_box, \
    scale_coords, xyxy2xywh
from y5utils.plots import Annotator, colors
from y5utils.torch_utils import time_sync

LOGGER = logging.getLogger(__name__)


def autopad(k, p=None):  # kernel, padding
    # Pad to 'same'
    if p is None:
        p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
    return p


class Conv(nn.Module):
    # Standard convolution
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, act=True):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p), groups=g, bias=False)
        self.bn = nn.BatchNorm2d(c2)
        self.act = nn.SiLU() if act is True else (act if isinstance(act, nn.Module) else nn.Identity())

    def forward(self, x):
        return self.act(self.bn(self.conv(x)))

    def forward_fuse(self, x):
        return self.act(self.conv(x))


class DWConv(Conv):
    # Depth-wise convolution class
    def __init__(self, c1, c2, k=1, s=1, act=True):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__(c1, c2, k, s, g=math.gcd(c1, c2), act=act)


class TransformerLayer(nn.Module):
    # Transformer layer https://arxiv.org/abs/2010.11929 (LayerNorm layers removed for better performance)
    def __init__(self, c, num_heads):
        super().__init__()
        self.q = nn.Linear(c, c, bias=False)
        self.k = nn.Linear(c, c, bias=False)
        self.v = nn.Linear(c, c, bias=False)
        self.ma = nn.MultiheadAttention(embed_dim=c, num_heads=num_heads)
        self.fc1 = nn.Linear(c, c, bias=False)
        self.fc2 = nn.Linear(c, c, bias=False)

    def forward(self, x):
        x = self.ma(self.q(x), self.k(x), self.v(x))[0] + x
        x = self.fc2(self.fc1(x)) + x
        return x


class TransformerBlock(nn.Module):
    # Vision Transformer https://arxiv.org/abs/2010.11929
    def __init__(self, c1, c2, num_heads, num_layers):
        super().__init__()
        self.conv = None
        if c1 != c2:
            self.conv = Conv(c1, c2)
        self.linear = nn.Linear(c2, c2)  # learnable position embedding
        self.tr = nn.Sequential(*[TransformerLayer(c2, num_heads) for _ in range(num_layers)])
        self.c2 = c2

    def forward(self, x):
        if self.conv is not None:
            x = self.conv(x)
        b, _, w, h = x.shape
        p = x.flatten(2).unsqueeze(0).transpose(0, 3).squeeze(3)
        return self.tr(p + self.linear(p)).unsqueeze(3).transpose(0, 3).reshape(b, self.c2, w, h)


class Bottleneck(nn.Module):
    # Standard bottleneck
    def __init__(self, c1, c2, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_, c2, 3, 1, g=g)
        self.add = shortcut and c1 == c2

    def forward(self, x):
        return x + self.cv2(self.cv1(x)) if self.add else self.cv2(self.cv1(x))


class BottleneckCSP(nn.Module):
    # CSP Bottleneck https://github.com/WongKinYiu/CrossStagePartialNetworks
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = nn.Conv2d(c1, c_, 1, 1, bias=False)
        self.cv3 = nn.Conv2d(c_, c_, 1, 1, bias=False)
        self.cv4 = Conv(2 * c_, c2, 1, 1)
        self.bn = nn.BatchNorm2d(2 * c_)  # applied to cat(cv2, cv3)
        self.act = nn.LeakyReLU(0.1, inplace=True)
        self.m = nn.Sequential(*[Bottleneck(c_, c_, shortcut, g, e=1.0) for _ in range(n)])

    def forward(self, x):
        y1 = self.cv3(self.m(self.cv1(x)))
        y2 = self.cv2(x)
        return self.cv4(self.act(self.bn(torch.cat((y1, y2), dim=1))))


class C3(nn.Module):
    # CSP Bottleneck with 3 convolutions
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c1, c_, 1, 1)
        self.cv3 = Conv(2 * c_, c2, 1)  # act=FReLU(c2)
        self.m = nn.Sequential(*[Bottleneck(c_, c_, shortcut, g, e=1.0) for _ in range(n)])
        # self.m = nn.Sequential(*[CrossConv(c_, c_, 3, 1, g, 1.0, shortcut) for _ in range(n)])

    def forward(self, x):
        return self.cv3(torch.cat((self.m(self.cv1(x)), self.cv2(x)), dim=1))


class C3TR(C3):
    # C3 module with TransformerBlock()
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):
        super().__init__(c1, c2, n, shortcut, g, e)
        c_ = int(c2 * e)
        self.m = TransformerBlock(c_, c_, 4, n)


class C3SPP(C3):
    # C3 module with SPP()
    def __init__(self, c1, c2, k=(5, 9, 13), n=1, shortcut=True, g=1, e=0.5):
        super().__init__(c1, c2, n, shortcut, g, e)
        c_ = int(c2 * e)
        self.m = SPP(c_, c_, k)


class C3Ghost(C3):
    # C3 module with GhostBottleneck()
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):
        super().__init__(c1, c2, n, shortcut, g, e)
        c_ = int(c2 * e)  # hidden channels
        self.m = nn.Sequential(*[GhostBottleneck(c_, c_) for _ in range(n)])


class SPP(nn.Module):
    # Spatial Pyramid Pooling (SPP) layer https://arxiv.org/abs/1406.4729
    def __init__(self, c1, c2, k=(5, 9, 13)):
        super().__init__()
        c_ = c1 // 2  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_ * (len(k) + 1), c2, 1, 1)
        self.m = nn.ModuleList([nn.MaxPool2d(kernel_size=x, stride=1, padding=x // 2) for x in k])

    def forward(self, x):
        x = self.cv1(x)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')  # suppress torch 1.9.0 max_pool2d() warning
            return self.cv2(torch.cat([x] + [m(x) for m in self.m], 1))


class SPPF(nn.Module):
    # Spatial Pyramid Pooling - Fast (SPPF) layer for YOLOv5 by Glenn Jocher
    def __init__(self, c1, c2, k=5):  # equivalent to SPP(k=(5, 9, 13))
        super().__init__()
        c_ = c1 // 2  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_ * 4, c2, 1, 1)
        self.m = nn.MaxPool2d(kernel_size=k, stride=1, padding=k // 2)

    def forward(self, x):
        x = self.cv1(x)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')  # suppress torch 1.9.0 max_pool2d() warning
            y1 = self.m(x)
            y2 = self.m(y1)
            return self.cv2(torch.cat([x, y1, y2, self.m(y2)], 1))


class Focus(nn.Module):
    # Focus wh information into c-space
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, act=True):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.conv = Conv(c1 * 4, c2, k, s, p, g, act)
        # self.contract = Contract(gain=2)

    def forward(self, x):  # x(b,c,w,h) -> y(b,4c,w/2,h/2)
        return self.conv(torch.cat([x[..., ::2, ::2], x[..., 1::2, ::2], x[..., ::2, 1::2], x[..., 1::2, 1::2]], 1))
        # return self.conv(self.contract(x))


class GhostConv(nn.Module):
    # Ghost Convolution https://github.com/huawei-noah/ghostnet
    def __init__(self, c1, c2, k=1, s=1, g=1, act=True):  # ch_in, ch_out, kernel, stride, groups
        super().__init__()
        c_ = c2 // 2  # hidden channels
        self.cv1 = Conv(c1, c_, k, s, None, g, act)
        self.cv2 = Conv(c_, c_, 5, 1, None, c_, act)

    def forward(self, x):
        y = self.cv1(x)
        return torch.cat([y, self.cv2(y)], 1)


class GhostBottleneck(nn.Module):
    # Ghost Bottleneck https://github.com/huawei-noah/ghostnet
    def __init__(self, c1, c2, k=3, s=1):  # ch_in, ch_out, kernel, stride
        super().__init__()
        c_ = c2 // 2
        self.conv = nn.Sequential(GhostConv(c1, c_, 1, 1),  # pw
                                  DWConv(c_, c_, k, s, act=False) if s == 2 else nn.Identity(),  # dw
                                  GhostConv(c_, c2, 1, 1, act=False))  # pw-linear
        self.shortcut = nn.Sequential(DWConv(c1, c1, k, s, act=False),
                                      Conv(c1, c2, 1, 1, act=False)) if s == 2 else nn.Identity()

    def forward(self, x):
        return self.conv(x) + self.shortcut(x)


class Contract(nn.Module):
    # Contract width-height into channels, i.e. x(1,64,80,80) to x(1,256,40,40)
    def __init__(self, gain=2):
        super().__init__()
        self.gain = gain

    def forward(self, x):
        b, c, h, w = x.size()  # assert (h / s == 0) and (W / s == 0), 'Indivisible gain'
        s = self.gain
        x = x.view(b, c, h // s, s, w // s, s)  # x(1,64,40,2,40,2)
        x = x.permute(0, 3, 5, 1, 2, 4).contiguous()  # x(1,2,2,64,40,40)
        return x.view(b, c * s * s, h // s, w // s)  # x(1,256,40,40)


class Expand(nn.Module):
    # Expand channels into width-height, i.e. x(1,64,80,80) to x(1,16,160,160)
    def __init__(self, gain=2):
        super().__init__()
        self.gain = gain

    def forward(self, x):
        b, c, h, w = x.size()  # assert C / s ** 2 == 0, 'Indivisible gain'
        s = self.gain
        x = x.view(b, s, s, c // s ** 2, h, w)  # x(1,2,2,16,80,80)
        x = x.permute(0, 3, 4, 1, 5, 2).contiguous()  # x(1,16,80,2,80,2)
        return x.view(b, c // s ** 2, h * s, w * s)  # x(1,16,160,160)


class Concat(nn.Module):
    # Concatenate a list of tensors along dimension
    def __init__(self, dimension=1):
        super().__init__()
        self.d = dimension

    def forward(self, x):
        return torch.cat(x, self.d)


class AutoShape(nn.Module):
    # YOLOv5 input-robust model wrapper for passing cv2/np/PIL/torch inputs. Includes preprocessing, inference and NMS
    conf = 0.25  # NMS confidence threshold
    iou = 0.45  # NMS IoU threshold
    classes = None  # (optional list) filter by class
    multi_label = False  # NMS multiple labels per box
    max_det = 1000  # maximum number of detections per image

    def __init__(self, model):
        super().__init__()
        self.model = model.eval()

    def autoshape(self):
        LOGGER.info('AutoShape already enabled, skipping... ')  # model already converted to model.autoshape()
        return self

    @torch.no_grad()
    def forward(self, imgs, size=640, augment=False, profile=False):
        # Inference from various sources. For height=640, width=1280, RGB images example inputs are:
        #   file:       imgs = 'data/images/zidane.jpg'  # str or PosixPath
        #   URI:             = 'https://ultralytics.com/images/zidane.jpg'
        #   OpenCV:          = cv2.imread('image.jpg')[:,:,::-1]  # HWC BGR to RGB x(640,1280,3)
        #   PIL:             = Image.open('image.jpg') or ImageGrab.grab()  # HWC x(640,1280,3)
        #   numpy:           = np.zeros((640,1280,3))  # HWC
        #   torch:           = torch.zeros(16,3,320,640)  # BCHW (scaled to size=640, 0-1 values)
        #   multiple:        = [Image.open('image1.jpg'), Image.open('image2.jpg'), ...]  # list of images

        t = [time_sync()]
        p = next(self.model.parameters())  # for device and type
        if isinstance(imgs, torch.Tensor):  # torch
            with amp.autocast(enabled=p.device.type != 'cpu'):
                return self.model(imgs.to(p.device).type_as(p), augment, profile)  # inference

        # Pre-process
        n, imgs = (len(imgs), imgs) if isinstance(imgs, list) else (1, [imgs])  # number of images, list of images
        shape0, shape1, files = [], [], []  # image and inference shapes, filenames
        for i, im in enumerate(imgs):
            f = f'image{i}'  # filename
            if isinstance(im, (str, Path)):  # filename or uri
                im, f = Image.open(requests.get(im, stream=True).raw if str(im).startswith('http') else im), im
                im = np.asarray(exif_transpose(im))
            elif isinstance(im, Image.Image):  # PIL Image
                im, f = np.asarray(exif_transpose(im)), getattr(im, 'filename', f) or f
            files.append(Path(f).with_suffix('.jpg').name)
            if im.shape[0] < 5:  # image in CHW
                im = im.transpose((1, 2, 0))  # reverse dataloader .transpose(2, 0, 1)
            im = im[..., :3] if im.ndim == 3 else np.tile(im[..., None], 3)  # enforce 3ch input
            s = im.shape[:2]  # HWC
            shape0.append(s)  # image shape
            g = (size / max(s))  # gain
            shape1.append([y * g for y in s])
            imgs[i] = im if im.data.contiguous else np.ascontiguousarray(im)  # update
        shape1 = [make_divisible(x, int(self.stride.max())) for x in np.stack(shape1, 0).max(0)]  # inference shape
        x = [letterbox(im, new_shape=shape1, auto=False)[0] for im in imgs]  # pad
        x = np.stack(x, 0) if n > 1 else x[0][None]  # stack
        x = np.ascontiguousarray(x.transpose((0, 3, 1, 2)))  # BHWC to BCHW
        x = torch.from_numpy(x).to(p.device).type_as(p) / 255.  # uint8 to fp16/32
        t.append(time_sync())

        with amp.autocast(enabled=p.device.type != 'cpu'):
            # Inference
            y = self.model(x, augment, profile)[0]  # forward
            t.append(time_sync())

            # Post-process
            y = non_max_suppression(y, self.conf, iou_thres=self.iou, classes=self.classes,
                                    multi_label=self.multi_label, max_det=self.max_det)  # NMS
            for i in range(n):
                scale_coords(shape1, y[i][:, :4], shape0[i])

            t.append(time_sync())
            return Detections(imgs, y, files, t, self.names, x.shape)


class Detections:
    # YOLOv5 detections class for inference results
    def __init__(self, imgs, pred, files, times=None, names=None, shape=None):
        super().__init__()
        d = pred[0].device  # device
        gn = [torch.tensor([*[im.shape[i] for i in [1, 0, 1, 0]], 1., 1.], device=d) for im in imgs]  # normalizations
        self.imgs = imgs  # list of images as numpy arrays
        self.pred = pred  # list of tensors pred[0] = (xyxy, conf, cls)
        self.names = names  # class names
        self.files = files  # image filenames
        self.xyxy = pred  # xyxy pixels
        self.xywh = [xyxy2xywh(x) for x in pred]  # xywh pixels
        self.xyxyn = [x / g for x, g in zip(self.xyxy, gn)]  # xyxy normalized
        self.xywhn = [x / g for x, g in zip(self.xywh, gn)]  # xywh normalized
        self.n = len(self.pred)  # number of images (batch size)
        self.t = tuple((times[i + 1] - times[i]) * 1000 / self.n for i in range(3))  # timestamps (ms)
        self.s = shape  # inference BCHW shape

    def display(self, pprint=False, show=False, save=False, crop=False, render=False, save_dir=Path('')):
        crops = []
        for i, (im, pred) in enumerate(zip(self.imgs, self.pred)):
            s = f'image {i + 1}/{len(self.pred)}: {im.shape[0]}x{im.shape[1]} '  # string
            if pred.shape[0]:
                for c in pred[:, -1].unique():
                    n = (pred[:, -1] == c).sum()  # detections per class
                    s += f"{n} {self.names[int(c)]}{'s' * (n > 1)}, "  # add to string
                if show or save or render or crop:
                    annotator = Annotator(im, example=str(self.names))
                    for *box, conf, cls in reversed(pred):  # xyxy, confidence, class
                        label = f'{self.names[int(cls)]} {conf:.2f}'
                        if crop:
                            file = save_dir / 'crops' / self.names[int(cls)] / self.files[i] if save else None
                            crops.append({'box': box, 'conf': conf, 'cls': cls, 'label': label,
                                          'im': save_one_box(box, im, file=file, save=save)})
                        else:  # all others
                            annotator.box_label(box, label, color=colors(cls))
                    im = annotator.im
            else:
                s += '(no detections)'

            im = Image.fromarray(im.astype(np.uint8)) if isinstance(im, np.ndarray) else im  # from np
            if pprint:
                LOGGER.info(s.rstrip(', '))
            if show:
                im.show(self.files[i])  # show
            if save:
                f = self.files[i]
                im.save(save_dir / f)  # save
                if i == self.n - 1:
                    LOGGER.info(f"Saved {self.n} image{'s' * (self.n > 1)} to {colorstr('bold', save_dir)}")
            if render:
                self.imgs[i] = np.asarray(im)
        if crop:
            if save:
                LOGGER.info(f'Saved results to {save_dir}\n')
            return crops

    def print(self):
        self.display(pprint=True)  # print results
        LOGGER.info(f'Speed: %.1fms pre-process, %.1fms inference, %.1fms NMS per image at shape {tuple(self.s)}' %
                    self.t)

    def show(self):
        self.display(show=True)  # show results

    def save(self, save_dir='runs/detect/exp'):
        save_dir = increment_path(save_dir, exist_ok=save_dir != 'runs/detect/exp', mkdir=True)  # increment save_dir
        self.display(save=True, save_dir=save_dir)  # save results

    def crop(self, save=True, save_dir='runs/detect/exp'):
        save_dir = increment_path(save_dir, exist_ok=save_dir != 'runs/detect/exp', mkdir=True) if save else None
        return self.display(crop=True, save=save, save_dir=save_dir)  # crop results

    def render(self):
        self.display(render=True)  # render results
        return self.imgs

    def pandas(self):
        # return detections as pandas DataFrames, i.e. print(results.pandas().xyxy[0])
        new = copy(self)  # return copy
        ca = 'xmin', 'ymin', 'xmax', 'ymax', 'confidence', 'class', 'name'  # xyxy columns
        cb = 'xcenter', 'ycenter', 'width', 'height', 'confidence', 'class', 'name'  # xywh columns
        for k, c in zip(['xyxy', 'xyxyn', 'xywh', 'xywhn'], [ca, ca, cb, cb]):
            a = [[x[:5] + [int(x[5]), self.names[int(x[5])]] for x in x.tolist()] for x in getattr(self, k)]  # update
            setattr(new, k, [pd.DataFrame(x, columns=c) for x in a])
        return new

    def tolist(self):
        # return a list of Detections objects, i.e. 'for result in results.tolist():'
        x = [Detections([self.imgs[i]], [self.pred[i]], self.names, self.s) for i in range(self.n)]
        for d in x:
            for k in ['imgs', 'pred', 'xyxy', 'xyxyn', 'xywh', 'xywhn']:
                setattr(d, k, getattr(d, k)[0])  # pop out of list
        return x

    def __len__(self):
        return self.n


class Classify(nn.Module):
    # Classification head, i.e. x(b,c1,20,20) to x(b,c2)
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.aap = nn.AdaptiveAvgPool2d(1)  # to x(b,c1,1,1)
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p), groups=g)  # to x(b,c2,1,1)
        self.flat = nn.Flatten()

    def forward(self, x):
        z = torch.cat([self.aap(y) for y in (x if isinstance(x, list) else [x])], 1)  # cat if list
        return self.flat(self.conv(z))  # flatten to x(b,c2)



"""
TODO: delete
"""
import os
from naeural_core import constants as ct


class _PersistenceSerializationMixin(object):
  def __init__(self):
    self.__persistence_fld = None
    super(_PersistenceSerializationMixin, self).__init__()
    return
  
  
  @property
  def _persistence_fld(self):
    return self.__persistence_fld
  
  
  @property
  def _cache_folder(self):
    return "" ## will save in data-root folder
  
  
  @property
  def plugin_id(self):
    """
    Returns the instance id of the current plugin.
    WARNING: This should be overwridden in the plugin class to return the correct id.

    Returns
    -------
    str
      the instance id.
      
    Example
    -------
      ```      
      instance_id = self.instance_id
      ```    
    """
    str_name = self.__class__.__name__
    str_name = self.sanitize_name(str_name)
    return str_name  


  def _maybe_setup_persistence_fld(self):
    if self.__persistence_fld is not None:
      return self.__persistence_fld       
    self.__persistence_fld = '{}/{}'.format(self._cache_folder, self.plugin_id.lower())
    full_path = os.path.join(self.log.get_data_folder(), self.__persistence_fld)
    os.makedirs(full_path, exist_ok=True)
    return self.__persistence_fld


  def _helper_persistence_serialization_save(self, obj, subfolder_path=None, verbose=True):
    self._maybe_setup_persistence_fld()
    if subfolder_path is None:
      subfolder_path = self._persistence_fld

    self.log.save_pickle_to_data(
      data=obj,
      fn='object.pickle',
      subfolder_path=subfolder_path,
      verbose=verbose
    )

    self.log.save_data_json(
      data_json={'SERIALIZATION_SIGNATURE' : self.cfg_serialization_signature},
      fname='metadata.json',
      subfolder_path=subfolder_path,
      verbose=verbose
    )
    return
    
  def _helper_persistence_serialization_load(self, subfolder_path=None, default=None, verbose=True):
    self._maybe_setup_persistence_fld()
    if subfolder_path is None:
      subfolder_path = self._persistence_fld

    dct_metadata = self.log.load_data_json(
      fname='metadata.json',
      subfolder_path=subfolder_path,
      verbose=verbose
    ) or {}

    loaded_serialization_signature = dct_metadata.get('SERIALIZATION_SIGNATURE', None)

    if self.cfg_serialization_signature is not None:
      if loaded_serialization_signature != self.cfg_serialization_signature:
        self.P("Loaded serialization signature differs from the current one", color='y')
        return default

    obj = self.log.load_pickle_from_data(
      fn='object.pickle',
      subfolder_path=subfolder_path,
      verbose=verbose
    ) or default

    return obj

  def _helper_persistence_serialization_update(self, update_callback, subfolder_path=None, verbose=True):
    self._maybe_setup_persistence_fld()
    if subfolder_path is None:
      subfolder_path = self._persistence_fld

    self.log.update_pickle_from_data(
      fn='object.pickle',
      update_callback=update_callback,
      subfolder_path=subfolder_path,
      verbose=verbose,
      force_update=True
    )
    self.log.save_data_json(
      data_json={'SERIALIZATION_SIGNATURE' : self.cfg_serialization_signature},
      fname='metadata.json',
      subfolder_path=subfolder_path,
      verbose=verbose
    )
    return
  
  def _helper_persistence_serialization_save_json(self, obj, subfolder_path=None, verbose=True):
    self._maybe_setup_persistence_fld()
    if subfolder_path is None:
      subfolder_path = self._persistence_fld

    self.log.save_data_json(
      data_json=obj,
      fname='data.json',
      subfolder_path=subfolder_path,
      verbose=verbose
    )
    return  
  
  def _helper_persistence_serialization_load_json(self, subfolder_path=None, default={}, verbose=True):
    self._maybe_setup_persistence_fld()
    if subfolder_path is None:
      subfolder_path = self._persistence_fld

    data = self.log.load_data_json(
      fname='data.json',
      subfolder_path=subfolder_path,
      verbose=verbose
    ) or default

    return data  
  
  
  
  def persistence_serialization_save(self, obj, verbose=True):
    self._helper_persistence_serialization_save(
      obj=obj,
      verbose=verbose
    )
    return

  def persistence_serialization_save_to_serving(self, obj, name, verbose=True):
    subfolder_path = os.path.join(ct.CACHE_SERVING, name.lower())
    self._helper_persistence_serialization_save(
      obj=obj,
      subfolder_path=subfolder_path,
      verbose=verbose
    )
    return

  def persistence_serialization_load(self, default=None, verbose=True):
    return self._helper_persistence_serialization_load(
      default=default,
      verbose=verbose
    )

  def persistence_serialization_load_from_serving(self, name, default=None, verbose=True):
    subfolder_path = os.path.join(ct.CACHE_SERVING, name.lower())
    return self._helper_persistence_serialization_load(
      subfolder_path=subfolder_path,
      default=default,
      verbose=verbose
    )

  def persistence_serialization_update(self, update_callback, verbose=True):
    self._helper_persistence_serialization_update(
      update_callback=update_callback,
      verbose=verbose
    )
    return

  def persistence_serialization_update_serving(self, update_callback, name, verbose=True):
    subfolder_path = os.path.join(ct.CACHE_SERVING, name.lower())
    self._helper_persistence_serialization_update(
      update_callback=update_callback,
      subfolder_path=subfolder_path,
      verbose=verbose
    )
    return
  
  
  def cacheapi_save_pickle(self, obj, verbose=True):
    """
    Save object to the current plugin instance cache folder

    Parameters
    ----------
    obj : any
        the picklable object to be saved
    verbose : bool, optional
        show information during process, by default True
    """
    self.persistence_serialization_save(obj, verbose=verbose)
    return
  
  
  def cacheapi_load_pickle(self, default=None, verbose=True):
    """
    Loads object from the current plugin instance cache folder

    Parameters
    ----------
    default : any, optional
        default value, by default None
    verbose : bool, optional
         show information during process, by default True

    Returns
    -------
    any
        the loaded object
    """
    return self.persistence_serialization_load(default=default, verbose=verbose)
  
  
  
  def cacheapi_save_json(self, obj, verbose=False):
    """
    Save object json to the current plugin instance cache folder

    Parameters
    ----------
    obj : any
        the json-able object to be saved
        
    verbose : bool, optional
        show information during process, by default True
    """
    self._helper_persistence_serialization_save_json(obj, verbose=verbose)
    return
  
  def cacheapi_load_json(self, default={}, verbose=False):
    """
    Loads object json from the current plugin instance cache folder

    Parameters
    ----------
    default : any, optional
        default value, by default {}
    verbose : bool, optional
         show information during process, by default True

    Returns
    -------
    any
        the loaded object
    """
    return self._helper_persistence_serialization_load_json(default=default, verbose=verbose)
  




if __name__ == '__main__':

  from naeural_core import Logger
  from naeural_core import DecentrAIObject

  log = Logger(lib_name='SER', base_folder='.', app_folder='_local_cache', TF_KERAS=False)

  class Base(DecentrAIObject, _PersistenceSerializationMixin):

    def __init__(self, load_prev_ser, ser_sign=None, **kwargs):
      self._load_prev_ser = load_prev_ser
      self._ser_sign = ser_sign
      self._instance_config = None
      super(Base, self).__init__(**kwargs)

    @property
    def unique_identification(self):
      return self._instance_config['STREAM'], self._instance_config['SIGNATURE'], self._instance_config['INSTANCE_ID']

    def startup(self):
      super().startup()
      self._instance_config = {
        'STREAM' : 'buc',
        'SIGNATURE' : 'camera_tampering_02',
        'INSTANCE_ID' : 'buc_ct_001',
        'LOAD_PREVIOUS_SERIALIZATION' : self._load_prev_ser,
        'SERIALIZATION_SIGNATURE' : self._ser_sign
      }
      self.P("In base startup")


  class Plu(Base):

    def __init__(self, **kwargs):
      super(Plu, self).__init__(**kwargs)
      self.obj = self.persistence_serialization_load()
      return

    def startup(self):
      super().startup()


  ##### Mode I - load if exists
  p = Plu(log=log, load_prev_ser=True, ser_sign=None)
  p.persistence_serialization_save({'a' : 1, 'b' : 2})
  log.P("\n")

  p = Plu(log=log, load_prev_ser=True, ser_sign=None)
  p.persistence_serialization_save({'a' : 100, 'b' : 100})
  log.P("\n")

  p = Plu(log=log, load_prev_ser=True, ser_sign='2123')
  p.persistence_serialization_save({'c': 1, 'd' : 2})
  log.P("\n")

  p = Plu(log=log, load_prev_ser=True, ser_sign='2123')
  log.P("\n")

  p = Plu(log=log, load_prev_ser=True, ser_sign=None)
  p.persistence_serialization_save({'c' : 100, 'd': 100})
  log.P("\n")

  p = Plu(log=log, load_prev_ser=True, ser_sign='2123')
  log.P("\n")

  p = Plu(log=log, load_prev_ser=True, ser_sign=None)
  log.P("\n")



import numpy as np
import itertools
from collections import OrderedDict
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

__VER__ = '0.1.1.0'

def get_basic_random_series(nr_series=50000, nr_points=750, fake_preds=False, noise_max=None):
  """
  Random series generator

  Parameters
  ----------
  nr_series : int, optional
    number of series. The default is 50000.
  nr_points : int, optional
    number of steps in each serie. The default is 750.
  fake_preds : bool, optional
    return also dummy predictions. The default is False.
  noise_max : bool, optional
    put a predefined max noise. The default is None.

  Returns
  -------
  res : ndarray
    the generated series.

  """
  if nr_series is None:
    nr_series = 50000
  funcs = [np.cos, np.sin]  
  x_signals = []
  y_signals = []
  fakes = []
  if noise_max is None:
    fake_delta = 1
  else:
    fake_delta = noise_max
  outlier_prc = 0.25
  funnels = []
  print_every = 1000 if nr_series >= 10000 else None
  for i in range(nr_series):
    step = np.random.choice(np.arange(0.01, 0.4, step=0.01),size=1)
    func = funcs[np.random.randint(len(funcs))]
    start = np.random.randint(10)
    np_base = np.arange(start,(start+1000)*np.pi,step)[:nr_points]
    np_outl = np.random.rand(nr_points)
    np_outl = (np_outl > (1 - outlier_prc)).astype(int)
    np_outv = np.random.uniform(fake_delta, 2*fake_delta, size=nr_points)    
    np_outv *= np_outl
    np_base = func(np_base) + 2*fake_delta
    np_fake = np.zeros((nr_points, 2))
    np_fake += np_base.reshape((nr_points,1)) - fake_delta
    np_fake[:,1] += fake_delta*2
    if noise_max is None:
      noise_size = np.random.rand()
    else:
      noise_size = np.random.uniform(0, noise_max)
    np_signal = np_base + np.random.uniform(-noise_size,noise_size, size=np_base.shape) #np.random.rand(*np_base.shape)
    np_signal += np_outv
    y_signals.append(np_signal.reshape((-1,1)))  
    x_signals.append(np.array([0] + np_signal[:-1].tolist()).reshape((-1,1)))
    fakes.append(np_fake)
    funnels.append(np.repeat(i, nr_points))
    if print_every is not None and (i % print_every) == 0:
      print("\rGenerating series {:.1f}%\r".format((i+1)/nr_series * 100), flush=True, end="")
  print("\rGenerating ndarrays...", flush=True, end="")  
  yt = np.array(y_signals)  
  x_feat = np.array(x_signals)
  x_funnel = np.array(funnels)
  ypf = np.array(fakes)
  res = (x_feat, x_funnel, yt, ypf) if fake_preds else (x_feat, x_funnel, yt)
  print("\rDone generating pseudo-series...\r", flush=True, end="")  
  return res


def get_random_series_batch(periods, n_points, random_start=True):
  if type(periods) not in [list, np.ndarray]:
    raise ValueError("periods must be a vector or list(nr_series) with all series periods")
  np_periods = np.array(periods).reshape((-1,1))
  n_series = np_periods.shape[0]
  np_steps = (2 * np.pi) / np_periods
  rnd_offset = np.random.randint(low=0, high=np_periods.max(), size=(n_series,1)) if random_start else 0 
  np_x = np.array([np.arange(0, n_points) for _ in range(len(periods))])
  np_x += rnd_offset
  np_x = np_x * np_steps
  np_y = np.sin(np_x)
  np_y = np_y + np_y.max(axis=1).reshape((-1,1))
  return np_y

def compose_series(np_series):
  lst_series_components = []
  for nr_series_composed in range(2, np_series.shape[0]+1):
    lst_series_components += list(itertools.combinations(list(range(np_series.shape[0])), nr_series_composed))

  lst_composed_series = []
  for components in lst_series_components:
    np_complex_series = np.zeros((np_series.shape[-1]))  
    for component in components:
      np_complex_series += np_series[component,:]
    lst_composed_series.append(np_complex_series)
  np_composed_series = np.array(lst_composed_series)
  
  return np_composed_series, lst_series_components
  
def generate_artificial_series(n_series, 
                               n_steps, 
                               freq, 
                               max_sales, 
                               encode_date_time=True,
                               noise_size=2,
                               random_peaks=False,
                               ):
  """
  TODO:
    - add normalization option for 0-1 target
    - add (semi) non-linear patterns
  
  This function generates a batch of time-series simulated sales.

  Parameters
  ----------
  n_series : int
    number of series.
    
  n_steps : int
    number of history steps.
    
  freq : str
    type of series: 'D' means daily, 'W' means weekly, 'M' means monthly.
    
  max_sales: int or np.ndarray
    max sales for all series (will generate a max for each ts) or list/array of maxes
        
  encode_date_time: bool
    True to return date covariate as float value (int part is year, fractional part is step (M, W, D))
    
    
  random_peaks: bool, optional
    The series will gradually increase and peak at random points. Default is disabled
    
    

  Returns
  -------
    data dictionary containing:
      ['SIGNAL'] : ndarray[n_series, n_steps] with the time-seriess
      ['COVAR'] : ndarray[n_series, n_steps] with the date covariate
      ['PERIODS'] : ndarray[n_series, 1] with periodicity of each signal

  """
  np_result = None
  RULES = OrderedDict({
    # daily
    'D': {
      'PERIODS' : [7, 30, 91, 182, 365],
      'MULT' : 1, # multiply factor to determine how many days based on n_steps
      'PER_STEP' : 0.001, # for date encoding is not start and step
      'PER_MAX'  : 0.366, # for date encoding is max of year period
      'PER_SIZE' : 365,   # for date encoding to see how many years
      'TIME_DELTA' : timedelta(days=1),
      }, 
    
    # weekly
    'W': {
      'PERIODS' : [4, 13, 26, 52],
      'MULT' : 7,
      'PER_STEP' : 0.01,
      'PER_MAX'  : 0.53,
      'PER_SIZE' : 52,   # for date encoding to see how many years
      'TIME_DELTA' : timedelta(weeks=1),
      }, 
    
    # monthly
    'M': {
      'PERIODS' : [6, 12],
      'MULT' : 30.5,
      'PER_STEP' : 0.01,
      'PER_MAX'  : 0.13,
      'PER_SIZE' : 12,   # for date encoding to see how many years
      'TIME_DELTA' : relativedelta(months=1),
      }  
    })
  lst_imp = list(RULES.keys())
  if freq not in lst_imp:
    raise ValueError("Time series frequency {} is not implemented. Valid options are {}".format(
      freq, lst_imp))
  freq = freq.upper()
  rules = RULES[freq[0]]
  periods = rules['PERIODS']
  
  # Composed series variants (including single component)
  lst_series_components = []
  for nr_series_composed in range(1, len(periods)):
    lst_series_components += list(itertools.combinations(periods, nr_series_composed))

  # now we generated the splines including composed periods
  
  np_series = get_random_series_batch(
    periods=periods,
    n_points=n_steps,
    random_start=True,
    )

  composed_series, lst_series_components = compose_series(np_series)
  
  np_series = np.concatenate((np_series, composed_series), axis=0)

  composed_periods = [[periods[y] for y in x] for x in lst_series_components]
  if True:
    periods = composed_periods

  np_series_indexes = np.random.choice(list(range(np_series.shape[0])), size=(n_series))
  
  np_series = np_series[np_series_indexes, :]
  periods = [periods[x] for x in np_series_indexes]
  # now we rescale the splines
  if type(max_sales) == np.ndarray:
    if max_sales.shape[0] != n_series:
      raise ValueError('Max sales {} must have same number of vectors as number of series {}'.format(
        max_sales.shape, n_series))
    np_max_sales = max_sales.copy()
    if len(np_max_sales.shape) == 1 or len(np_max_sales.shape[1]) == 1:
      np_max_sales = np_max_sales.reshape((-1, 1))
  else:
    np_max_sales = np.random.randint(low=1, high=max_sales, size=(n_series, 1))
  
  if random_peaks:
    peak_point = np.random.randint(n_steps // 2, n_steps // 2 + n_steps // 3)
    min_per_serie = np_max_sales.max(1) / 2
    max_per_serie = np_max_sales.max(1) * 1.5
    if np_max_sales.shape[1] != n_steps:      
      np_max_sales = np.concatenate(
        (np_max_sales, np.zeros((n_series, n_steps - np_max_sales.shape[1]))),
        axis=1
        )
    for idx_ser in range(n_series):
      np_max_sales[idx_ser] = np.concatenate((
        np.linspace(
          min_per_serie[idx_ser], 
          max_per_serie[idx_ser], 
          num=peak_point
          ),
        np.linspace(
          max_per_serie[idx_ser], 
          min_per_serie[idx_ser], 
          num=n_steps - peak_point
          ),
        ))
      
    
  np_rescaled = np_series * np_max_sales
  
  # finally we add noise, round to integers and constraint lower bound zero
  np_noisy = np_rescaled + np.random.randint(
    low=-noise_size, 
    high=noise_size, 
    size=(n_series, n_steps)
    )
  np_result = np.floor(np.clip(a=np_noisy, a_min=0, a_max=None))
  
  # now lets prepare the covariates
  np_dates = None
  if encode_date_time:
    per_start = RULES[freq]['PER_STEP']
    per_max = RULES[freq]['PER_MAX']
    nr_years = int(np.ceil(n_steps / RULES[freq]['PER_SIZE']))
    years = []
    for y in range(1, nr_years + 1):
      np_cy = np.arange(start=per_start, stop=per_max, step=per_start)
      np_cy = np_cy + y
      years.append(np_cy)
    np_serie_dates = np.concatenate(years)
  else:    
    curr_date = datetime.now()
    step_size = RULES[freq]['MULT']
    back_days = n_steps * step_size + 1
    start_date = curr_date.now().date() - timedelta(days=back_days)    
    td = RULES[freq]['TIME_DELTA']
    dates = [start_date + td * j for j in range(1, n_steps + 1)]
    np_serie_dates = np.array(dates)
  
  np_dates = np.array([np_serie_dates] * n_series)[:,:n_steps]
    
  
  result = {
    'SIGNAL' : np_result,
    'PERIODS' : periods,
    'COVAR' : np_dates,
    }
  
  return result

if __name__ == '__main__':
  import matplotlib.pyplot as plt
  np.set_printoptions(
    suppress=True,
    precision=2,
    floatmode='fixed',
    threshold=1000,
    linewidth=1000,
    )
  
  # n_points = 52
  # period = 4 + 4 + 5
  # step = 2 * np.pi / period
  # x = np.arange(0, n_points)
  # xi = x * step
  # y = np.sin(xi)
  # y += y.max()
  # plt.plot(x, y, linestyle='--', marker='o')
  # plt.title('BASIC')
  # plt.show()
  
  # periods = [4, 13, 26, 52]
  # np_y = get_random_series_batch(
  #   periods=periods,
  #   n_points=n_points,
  #   random_start=True
  #   )
  
  # for i in range(np_y.shape[0]):
  #   plt.plot(np.arange(0, n_points), np_y[i], linestyle='--', marker='+')
  #   plt.title('Series {} with period {}'.format(i, periods[i]))
  #   plt.show()
  
  data = generate_artificial_series(
    n_series=1,
    n_steps=1000,
    freq='D', 
    max_sales=10,
    encode_date_time=False,
    )
  np.random.seed(2)

  np_ser = data['SIGNAL']
  np_per = data['PERIODS']
  np_dates = data['COVAR']
  
  plot_params_1 = {
    'linestyle' : 'None',
    'marker'    : '+'
    }
  plot_params_2 = {
    'linestyle'       : '-',
    'marker'          : '+',
    'markeredgecolor' : 'red',
    'markeredgewidth' : 2,
    }
  plot_params = plot_params_2
  
  for i in range(np_ser.shape[0]):
    np_serie = np_ser[i]
    n_steps = np_serie.shape[0]
    # x = np.arange(0, np_ser.shape[1])
    x_labels = ['{}'.format(j) for j in np_dates[i]]
    plt.figure(figsize=(14,9))
    plt.plot(np_serie, **plot_params)
    #plt.xticks(np.arange(0, n_steps), labels=x_labels, rotation='vertical')
    plt.title('Series {} with periods {}'.format(i, np_per[i]))
    plt.show()
    
    
  


"""
TODO: 
   - merge with amqp controller implementation  - this is 1-to-1 unusefull separation
   as new plugins will have to implement the comm logic anyway
   - as mqtt is the "base" we can leave full amqp implementation here in user area
"""

from naeural_core.comm import AMQPWrapper
from naeural_core.comm.base import BaseCommThread

_CONFIG = {
  **BaseCommThread.CONFIG,
  'VALIDATION_RULES': {
    **BaseCommThread.CONFIG['VALIDATION_RULES'],
  },
}


class AMQPCommThread(BaseCommThread):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self._controller = None
    super(AMQPCommThread, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    return

  @property
  def send_to(self):
    return self._send_to

  @send_to.setter
  def send_to(self, x):
    self._send_to = x
    self._controller.send_channel_name = (self._send_channel_name, self._send_to)
    return

  @property
  def connection(self):
    return self._controller.connection

  @property
  def channel(self):
    if self._controller.channel is not None:
      if self._controller.channel.is_closed:
        return None

    return self._controller.channel

  def _get_server_address(self):
    return "{}:{}".format(self._controller.cfg_vhost, self._controller.cfg_port)

  # base comm abstract implementations

  def _init(self):
    self._controller = AMQPWrapper(
      log=self.log, config=self._config, recv_buff=self._recv_buff,
      send_channel_name=self._send_channel_name, recv_channel_name=self._recv_channel_name,
      comm_type=self._comm_type
    )
    self._maybe_reconnect_to_server()
    return

  def _maybe_reconnect_to_server(self):
    if self.connection is None or self.channel is None or not self.has_server_conn:
      # must reset otherwise we have same bug as in mqtt
      self.has_send_conn = False
      self.has_recv_conn = False
      dct_ret = self._controller.server_connect()
      self.has_server_conn = dct_ret['has_connection']
      msg = dct_ret['msg']
      msg_type = dct_ret['msg_type']
      self._create_notification(
        notif=msg_type,
        msg=msg
      )
      if not self.has_server_conn:
        self._total_conn_fails += 1
        self._error_messages.append((msg_type, msg))
        self._error_times.append(self.log.time_to_str())
        self.P(msg, color='r')
      else:
        self.P(msg)
    # endif
    return

  def _maybe_reconnect_recv(self):
    self._maybe_reconnect_to_server()
    if self.has_server_conn and not self.has_recv_conn:
      dct_ret = self._controller.establish_one_way_connection('recv')
      self.has_recv_conn = dct_ret['has_connection']
      msg = dct_ret['msg']
      msg_type = dct_ret['msg_type']
      self._create_notification(
        notif=msg_type,
        msg=msg
      )
      if not self.has_recv_conn:
        self._total_conn_fails += 1
        self._error_messages.append((msg_type, msg))
        self._error_times.append(self.log.time_to_str())
        self.P(msg, color='r')
      else:
        self.P(msg)
    # endif
    return

  def _maybe_reconnect_send(self):
    self._maybe_reconnect_to_server()
    if self.has_server_conn and not self.has_send_conn:
      dct_ret = self._controller.establish_one_way_connection('send')
      self.has_send_conn = dct_ret['has_connection']
      msg = dct_ret['msg']
      msg_type = dct_ret['msg_type']
      self._create_notification(
        notif=msg_type,
        msg=msg
      )
      if not self.has_send_conn:
        self._total_conn_fails += 1
        self._error_messages.append((msg_type, msg))
        self._error_times.append(self.log.time_to_str())
        self.P(msg, color='r')
      else:
        self.P(msg)
    # endif
    return

  def _maybe_fill_recv_buffer(self):
    self._controller.receive()
    return

  def _send(self, data, send_to=None):
    self._controller.send(data, send_to=send_to)
    return

  def _release(self):
    dct_ret = self._controller.release()
    for msg in dct_ret['msgs']:
      self.P(msg)

    self.has_send_conn = False
    self.has_recv_conn = False
    self.has_send_conn = False

    del self._controller
    return



"""
Javascript generated message:
r:     56684693624919999627438170883767084573186267748701669138919084275734263485571
s:     5787701224495155401849614702459149372951204779942523309061490057173504587799
Sign:  MEQCIH1SZk5vEQHxMKMKfh86BWtcv-9OxOJi9_YEHbNjMdSDAiAMy7kJzWki9Yg879ztF4KqDQYy4mOHn0RjYojaF1N4Fw==  
  
"""
import json
import base64

from hashlib import sha256
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.utils import decode_dss_signature

from naeural_core import Logger


if __name__ == '__main__':
  # now the received signature check process
  PYTHON_MESSAGE = {
    "9": 9,
    "2": 2,
    "3": 3,
    "10": {
      "2": 2,
      "100": 100,
      "1": 1
    },
    # "SIGN": "MEQCIHWA7Jmtz-ENKo3JfMbfWGGmIAufwCFkKXz35gV8VudyAiBV8KlIKRglYlH8odTF-yfwJvHDv28BioFc1Vm1OONJVw==",
    "SIGN": "MEQCIEIz_Nfy9CJ0GYW1V7Iw0uFJAVzu1TnOWkCVYnrt8PNHAiB0JCk_pgzGGIMz-KIvOCC_BzbGB8jxkAb_OwPX7AQTyA==",
    "ADDR": "0xai_AuN2SENcYNzRgbPUHVCFe6W1q-vieUKap2VY9mU_Fljy",
    "HASH": "7d72cf5bd6cda16c86dfb2c2c4983464edda5bf78e1ca3a21139f5037454c6f3"
  }
  
  JS_GENERATED_MESSAGE = {
    "2": 2,
    "3": 3,
    "9": 9,
    "10": {
      "1": 1,
      "2": 2,
      "100": 100
    },
    "SIGN": "MEQCIH1SZk5vEQHxMKMKfh86BWtcv-9OxOJi9_YEHbNjMdSDAiAMy7kJzWki9Yg879ztF4KqDQYy4mOHn0RjYojaF1N4Fw==",
    "ADDR": "0xai_AuN2SENcYNzRgbPUHVCFe6W1q-vieUKap2VY9mU_Fljy",
    "HASH": "7d72cf5bd6cda16c86dfb2c2c4983464edda5bf78e1ca3a21139f5037454c6f3"
  }
    
  JS_r = 56684693624919999627438170883767084573186267748701669138919084275734263485571
  JS_s = 5787701224495155401849614702459149372951204779942523309061490057173504587799
    
  NON_DATA_FIELDS = ['SIGN', 'ADDR', 'HASH']
  dct_received = JS_GENERATED_MESSAGE
  # dct_received['9'] = 10
  str_sender_addr = dct_received['ADDR']
  str_sender_sign = dct_received['SIGN']  
  str_sender_hash = dct_received['HASH']
  dct_data = {k : dct_received[k] for k in dct_received if k not in NON_DATA_FIELDS}
  str_deterministic_json = json.dumps(dct_data, sort_keys=True, separators=(',',':'))
  if str_deterministic_json == '{"10":{"1":1,"100":100,"2":2},"2":2,"3":3,"9":9}':
    Logger.print_color("Deterministic JSON passed", color='g')
  else:
    Logger.print_color("ERROR: Deterministic JSON failed", color='r')
    
  data = str_deterministic_json.encode()
  obj_digest = sha256(data)
  str_digest = obj_digest.hexdigest()
  bin_digest = obj_digest.digest()
  simple_addr = str_sender_addr.replace("0xai_", '')
  bpublic_key = base64.urlsafe_b64decode(simple_addr)
  pk = ec.EllipticCurvePublicKey.from_encoded_point(
    curve=ec.SECP256K1(), 
    data=bpublic_key
  )

  # RECODING: To get the hexadecimal representation of the public key
  public_bytes = pk.public_bytes(
      encoding=ec._serialization.Encoding.X962,
      format=ec._serialization.PublicFormat.CompressedPoint
  )  
  str_recoded_addr = base64.urlsafe_b64encode(public_bytes).decode()
  
  signature = base64.urlsafe_b64decode(str_sender_sign)  
  r, s = decode_dss_signature(signature)
  failed = False
  
  if dct_received == JS_GENERATED_MESSAGE:
    if r == JS_r and s == JS_s:
      Logger.print_color("Decoded signature r,s check passed", color='g')
    else:
      Logger.print_color("ERROR: Decoded signature r,s check failed", color='r')
      exit(1)
  else:
    print('r:     {}'.format(r))
    print('s:     {}'.format(s)) 
    print('Sign:  {}'.format(str_sender_sign))
      
  if str_recoded_addr != simple_addr:
    Logger.print_color("ERROR: Recoded address check failed", color='r')
    exit(1)
  else:
    Logger.print_color("Recoded address check passed", color='g')
    
  if str_digest != str_sender_hash:
    Logger.print_color("ERROR: Hash check failed", color='r')
    exit(1)
  else:
    Logger.print_color("Hash check passed", color='g')
    

  err = None
  try:
    pk.verify(signature, bin_digest, ec.ECDSA(hashes.SHA256()))
    signature_check = True
  except Exception as e:
    signature_check = False

  if signature_check:
    Logger.print_color("Signature check passed", color='g')
  else:
    Logger.print_color("ERROR: Signature check failed", color='r')    
